# imports
import typer

# typer config
## default kwargs
default_kwargs = {
    "no_args_is_help": True,
    "add_completion": False,
    "context_settings": {"help_option_names": ["-h", "--help"]},
}

## main app
app = typer.Typer(help="dkdc-todo", **default_kwargs)


# commands
## servers
@app.command()
@app.command("g", hidden=True)
def gui(
    port: int = typer.Option(8010, help="port", show_default=True),
    prod: bool = typer.Option(False, help="prod?", show_default=True),
):
    """
    gui
    """
    from shiny import run_app as run_gui_app
    from dkdc_todo.ui.gui import app  # noqa

    if prod:
        run_gui_app(
            app=app,
            host="0.0.0.0",
            port=port,
        )
    else:
        run_gui_app(
            app="dkdc_todo.ui.gui:app",  # goofy! but needed to reload
            host="0.0.0.0",
            port=port,
            reload=True,
            launch_browser=True,
        )


# main
if __name__ == "__main__":
    typer.run(app)



import os

from dkdc_util import now, uuid, get_dkdc_dir
from dkdc_state import State, ibis, dt


# class
class Todo(State):
    def __init__(self, dbpath: str = None):
        if dbpath is None:
            dbpath = os.path.join(get_dkdc_dir(), "todo.db")
        super().__init__(dbpath=dbpath)

    def _cons(self) -> (ibis.BaseBackend, ibis.BaseBackend):
        # create write connection
        wcon = ibis.sqlite.connect(self.dbpath)

        # create tables in write connection
        ## todos data
        self.todos_table_name = "todos"
        schema = ibis.schema(
            {
                "idx": dt.timestamp,
                "id": str,
                "user_id": str,
                "subject": str,
                "body": str,
                "priority": int,
                "status": str,
                "description": str,
                "labels": str,  # comma-separated list of labels
            }
        )
        if self.todos_table_name not in wcon.list_tables():
            wcon.create_table(self.todos_table_name, schema=schema)

        # create read connection
        rcon = ibis.duckdb.connect()

        # create tables in read connection
        for table_name in wcon.list_tables():
            rcon.read_sqlite(self.dbpath, table_name=table_name)

        return wcon, rcon

    # tables
    def todos_t(self, user_id: str = None):
        # get todos data
        t = self.rcon.table(self.todos_table_name)

        # filter by user_id
        if user_id:
            t = t.filter(t["user_id"] == user_id)

        # get only the latest metadata
        t = (
            t.mutate(
                rank=ibis.row_number().over(
                    ibis.window(
                        group_by="id",
                        order_by=ibis.desc("idx"),
                    )
                )
            )
            .filter(ibis._["rank"] == 0)
            .drop("rank")
        )

        # comma-separated lists to arrays
        t = t.mutate(labels=t["labels"].split(","))

        # order
        t = t.order_by(ibis.asc("priority"), ibis.desc("idx"))

        # return the data
        return t

    # contains
    def contains_todo(self, id: str, user_id: str = None) -> bool:
        t = self.todos_t(user_id=user_id)
        return t.filter(t["id"] == id).count().to_pyarrow().as_py() > 0

    # get record
    def get_todos(self, user_id: str = None):
        t = self.todos_t(user_id=user_id)

        return t.to_pyarrow().to_pylist()

    def get_todo(self, id: str, user_id: str = None):
        t = self.todos_t(user_id=user_id)
        return t.filter(t["id"] == id).to_pyarrow().to_pylist()[0]

    # append record
    def append_message(
        self,
        to: str,
        from_: str,
        convo_id: str,
        thread_id: str,
        subject: str,
        body: str,
        attachments: list[str] = None,
        version: int = None,
        status: str = None,
        description: str = None,
        labels: list[str] = None,
    ):
        assert (to is not None) and (from_ is not None), "to and from_ are required"

        data = {
            "idx": [now()],
            "id": [uuid()],
            "to": [to],
            "from": [from_],
            "convo_id": [convo_id],
            "thread_id": [thread_id],
            "subject": [subject],
            "body": [body],
            "attachments": [",".join(attachments) if attachments else None],
            "version": [version],
            "status": [status],
            "description": [description],
            "labels": [",".join(labels) if labels else None],
        }
        self.wcon.insert(self.todos_table_name, data)

        return self.get_todo(id=data["id"][0])

    def append_todo(
        self,
        id: str,
        user_id: str,
        subject: str,
        body: str,
        priority: int = 100,
        status: str = None,
        description: str = None,
        labels: list[str] = None,
    ):
        data = {
            "idx": [now()],
            "id": [id],
            "user_id": [user_id],
            "subject": [subject],
            "body": [body],
            "priority": [priority],
            "status": [status],
            "description": [description],
            "labels": [",".join(labels) if labels else None],
        }
        self.wcon.insert(self.todos_table_name, data)

        return self.get_todo(id=id)

    # update record
    def update_todo(
        self,
        id: str,
        user_id: str,
        subject: str,
        body: str,
        priority: int = 100,
        status: str = None,
        description: str = None,
        labels: list[str] = None,
    ):
        if not self.contains_todo(id=id):
            raise ValueError(f"todo {id} does not exist")

        data = {
            "idx": [now()],
            "id": [id],
            "user_id": [user_id],
            "subject": [subject],
            "body": [body],
            "priority": [priority],
            "status": [status],
            "description": [description],
            "labels": [",".join(labels) if labels else None],
        }
        self.wcon.insert(self.todos_table_name, data)

        return self.get_todo(id=id)

    # aliases
    __call__ = get_todos
    t = todos_t



# imports
from shiny import App, ui, reactive, render
from shinyswatch import theme

from dkdc_todo.ui.gui.todos import todo_page, todo_server

# gui
gui = ui.page_fluid(
    ui.output_ui("main"),
    theme=theme.pulse,
)


# server
def gui_server(input, output, session):
    # setup reactive values
    _val = reactive.Value(None)

    # WARNING: hack for testing only (so you don't have to login/use a test user)
    # current_username.set("test")

    # callbacks
    def _whatever(u):
        _val.set(u)

    # servers
    todo_server("todo")

    # global effects
    @render.ui
    def main():
        elements = [
            ui.nav_panel("home", ui.markdown("home")),
            ui.nav_panel("todo", todo_page("todo")),
        ]
        elements += [
            ui.nav_spacer(),
            ui.nav_control(
                ui.input_dark_mode(),
            ),
        ]

        return ui.navset_bar(
            *elements,
            title=ui.a("dkdc-io", href="https://dkdc.io", class_="navbar-brand"),
            id="navbar",
            selected="todo",
        )


# app
app = App(gui, gui_server)



# imports
from shiny import ui, render, reactive, module

from dkdc_todo import Todo
from dkdc_util import uuid_parts
from dkdc_state import ibis

# global state
todo = Todo()


# individual todo card
@module.ui
def todo_card(header, body, priority):
    return ui.card(
        ui.card_header(header),
        ui.layout_columns(
            ui.markdown(body),
            ui.input_slider("priority", "priority", value=priority, min=0, max=100),
        ),
        ui.layout_columns(
            ui.input_action_button("done", "done", class_="btn-primary"),
            ui.input_action_button("delete", "delete", class_="btn-danger"),
        ),
    )


@module.server
def todo_card_server(input, output, session, todos_modified):
    def _get_id(session):
        # TODO: this isn't ideal
        return str(session.ns).split("-")[-1]

    @reactive.Effect
    @reactive.event(input.priority, ignore_init=True)
    def _update_priority():
        id = _get_id(session)
        t = todo.get_todo(id=id)
        todo.update_todo(
            id=id,
            user_id=t["user_id"],
            subject=t["subject"],
            body=t["body"],
            priority=input.priority(),
            status=t["status"],
            description=t["description"],
            labels=t["labels"],
        )
        todos_modified()

    @reactive.Effect
    @reactive.event(input.done, ignore_init=True)
    def _done():
        id = _get_id(session)
        t = todo.get_todo(id=id)
        todo.update_todo(
            id=id,
            user_id=t["user_id"],
            subject=t["subject"],
            body=t["body"],
            priority=t["priority"],
            status="done",
            description=t["description"],
            labels=t["labels"],
        )
        todos_modified()

    @reactive.Effect
    @reactive.event(input.delete, ignore_init=True)
    def _delete():
        id = _get_id(session)
        t = todo.get_todo(id=id)
        todo.update_todo(
            id=id,
            user_id=t["user_id"],
            subject=t["subject"],
            body=t["body"],
            priority=t["priority"],
            status="deleted",
            description=t["description"],
            labels=t["labels"],
        )
        todos_modified()


# page of todos
@module.ui
def todo_page():
    return ui.page_fluid(
        ui.layout_columns(
            ui.card(
                ui.card_header("add todo"),
                ui.layout_columns(
                    ui.input_text("todo", "todo"),
                    ui.input_slider("priority", "priority", value=0, min=0, max=100),
                ),
                ui.layout_columns(
                    ui.input_action_button("add", "add", class_="btn-primary"),
                    ui.input_action_button("clear", "clear all", class_="btn-danger"),
                ),
            ),
            ui.card(
                ui.card_header("todos stats"),
                ui.output_ui("stats"),
            ),
        ),
        ui.card(
            ui.card_header("todos"),
            ui.output_ui("todos_list"),
        ),
    )


@module.server
def todo_server(input, output, session):
    # reactive values
    todo_modified = reactive.Value(0)

    # servers
    [
        todo_card_server(
            t["id"], todos_modified=lambda: todo_modified.set(todo_modified.get() + 1)
        )
        for t in todo.get_todos()
    ]

    # effects
    @render.ui
    def stats():
        _ = todo_modified.get()
        return ui.markdown(
            f"total todos: {todo.t().filter(ibis._["status"].isnull()).count().to_pyarrow().as_py()}"
        )

    @reactive.Effect
    @reactive.event(input.add, ignore_init=True)
    def _add():
        _, id = uuid_parts()
        todo_text = input.todo()
        todo_priority = input.priority()
        todo.append_todo(
            id=id,
            user_id=None,
            subject=None,
            body=todo_text,
            priority=todo_priority,
        )
        todo_card_server(
            id, todos_modified=lambda: todo_modified.set(todo_modified.get() + 1)
        )
        todo_modified.set(todo_modified.get() + 1)

    @reactive.Effect
    @reactive.event(input.clear, ignore_init=True)
    def _clear():
        for t in todo.t().filter(ibis._["status"].isnull()).to_pyarrow().to_pylist():
            todo.update_todo(
                id=t["id"],
                user_id=t["user_id"],
                subject=t["subject"],
                body=t["body"],
                priority=t["priority"],
                status="deleted",
                description=t["description"],
                labels=t["labels"],
            )
        todo_modified.set(todo_modified.get() + 1)

    @render.ui
    def todos_list():
        _ = todo_modified.get()
        return (
            ui.layout_column_wrap(
                *[
                    todo_card(t["id"], t["id"], t["body"], t["priority"])
                    for t in todo.t()
                    .filter(ibis._["status"].isnull())
                    .to_pyarrow()
                    .to_pylist()
                ]
            ),
        )
